from python_grains.dynamic_settings.lua import LuaScripts

import requests
from requests.sessions import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
import json
import datetime
import pytz
import redis
import os

DEFAULT_DEBUG_VERSION = 'QUICK_PREVIEW'
DEFAULT_MAX_AGE = datetime.timedelta(days=21)

class DynamicSetting(object):

    fallback_version = 'fallback'
    download_in_progress_ttl = 5
    data_margin_seconds = 24 * 60 * 60
    max_file_count = 3

    def __init__(self,
                 slug,
                 url,
                 redis_client,
                 domain,
                 logger,
                 data_dir,
                 fallback_data=None,
                 instance_id=None,
                 max_age=DEFAULT_MAX_AGE,
                 download_subkey=None,
                 version_subkey=None,
                 download_options=None,
                 debug_version=DEFAULT_DEBUG_VERSION,
                 fallback_allowed=False,
                 download_timeout=1.5,
                 verbose=False):

        self.slug = slug
        self.content = None
        self.version = None
        self.download_time = None
        self.load_time = None
        self.source = None
        self.fallback_allowed = fallback_allowed
        self.verbose = verbose

        self.domain = domain
        self.url = url
        self.download_subkey = download_subkey
        self.version_subkey = version_subkey
        self.download_options = download_options or {}

        self.max_age = max_age
        self.debug_version = debug_version
        self.redis_client = redis_client
        self.logger = logger
        self.instance_id = instance_id
        self.data_dir = data_dir
        self.download_try_time = None

        self.fallback_data = fallback_data or {}
        self.download_timeout = download_timeout
        self.data_ttl = self.max_age.total_seconds() + self.data_margin_seconds

        self.attach_scripts_to_redis_client()

    @classmethod
    def from_setting_object(cls,
                            obj,
                            redis_client,
                            domain,
                            data_dir,
                            logger=None,
                            instance_id=None,
                            verbose=False):

        return cls(slug=obj['slug'],
                   url=obj['url'],
                   redis_client=redis_client,
                   domain=domain,
                   logger=logger,
                   data_dir=data_dir,
                   fallback_data=obj['fallback_data'],
                   instance_id=instance_id,
                   max_age=obj['max_age'],
                   download_options=obj.get('download_options'),
                   download_subkey=obj.get('subkey'),
                   version_subkey=obj.get('version'),
                   fallback_allowed=obj.get('fallback_allowed', False),
                   download_timeout=obj.get('download_timeout', 1.5),
                   verbose=verbose)

    def get(self, version=None):

        if self.content is None or self.is_fallback:
            if self.verbose:
                print(f'{self.slug} - Loading from storage, since content is None or this is the fallback')
            self.load_from_storage(version)

        elif self.loaded_version_is_ok(version) and not self.too_old:
            if self.verbose:
                print(f'{self.slug} - Loaded content is ok')
            return self.content

        else:
            self.load_from_storage(version)

        if self.content is None:
            if self.verbose:
                print(f'{self.slug} - Loading fallback')
            self.load_fallback()

        return self.content

    def load_content(self, data, version, download_time, source):

        if hasattr(self, 'process_' + self.slug) and callable(getattr(self, 'process_' + self.slug)):
            data = getattr(self, 'process_' + self.slug)(data)

        self.content = data
        self.version = version
        self.download_time = download_time
        self.load_time = self.current_time
        self.source = source

    def load_from_storage(self, version):

        self.load_from_cache(version)

        if self.content is None or self.too_old:
            self.download()

    def load_from_cache(self, version):

        if self.verbose:
            print(f'{self.slug} - Loading from cache')
        try:
            data = self.redis_client.get_latest_data(
                keys=[
                    self.current_version_key
                ],
                args=[
                    self.data_key_prefix,
                    self.data_ttl
                ]
            )
            cache_version, download_time, content = self.parse_cache_data(data=data)
            source = 'cache'
            if self.verbose:
                print(f'{self.slug} - Loaded from cache and content is {"" if content is None else "not "}None')
        except (redis.connection.ConnectionError, redis.connection.TimeoutError):
            if self.verbose:
                print(f'{self.slug} - Loading from cache failed')
            cache_version, download_time, content = self.load_from_file()
            source = 'disk'


        if not content is None:

            if self.is_larger_than_loaded_version(cache_version):
                if self.verbose:
                    print(f'{self.slug} - Cache version larger than loaded version')
                self.load_content(data=content,
                                  version=cache_version,
                                  download_time=download_time,
                                  source=source)
            if self.loaded_version_is_ok(version):
                if self.verbose:
                    print(f'{self.slug} - Loaded version is ok')
                return

        self.download()

    def load_from_file(self):

        self.logger.warning('Loading settings from file', data=self.log_data)
        if self.verbose:
            print(f'{self.slug} - Loading from disk')
        latest_files = [fn for fn in [self.parse_filename(filename) for filename in os.listdir(self.data_dir)
                                      if filename.endswith('.json') and filename.startswith(self.slug)
                                      and not filename.endswith('fallback.json')] if not fn is None]
        if not latest_files:
            if self.verbose:
                print(f'{self.slug} - No file to load from')
            return None, None, None

        latest_file = sorted(latest_files, key=lambda x: int(x['version']))[-1]
        try:
            with open(latest_file['file_path'], 'r') as f:
                data = json.load(f)
                if self.verbose:
                    print(f'{self.slug} - Loaded from disk')
                return self.parse_data(data)
        except json.JSONDecodeError:
            if self.verbose:
                print(f'{self.slug} - Invalid file on disk')
            return None, None, None


    def parse_data(self, data):

        version = data['version']
        download_time = pytz.utc.localize(datetime.datetime.utcfromtimestamp(float(data['timestamp'])))
        content = data['content']

        return version, download_time, content

    def parse_cache_data(self, data):

        if data is None:
            return None, None, None

        data = data.decode('utf-8') if isinstance(data, bytes) else data
        data = json.loads(data)

        return self.parse_data(data=data)

    def load_fallback(self):

        self.load_content(data=self.fallback_data,
                          version=self.fallback_version,
                          download_time=None,
                          source='fallback')

    def is_debug_version(self, version):

        return str(version).upper() == self.debug_version

    def loaded_version_is_ok(self, version):

        if self.is_debug_version(version) or version is None:
            return True

        try:
            return int(self.version) >= int(version)
        except ValueError:
            return True

    def is_larger_than_loaded_version(self, version):

        if self.version is None or self.version == self.fallback_version:
            return True
        try:
            return int(version) > int(self.version)
        except ValueError:
            return False

    def attach_scripts_to_redis_client(self):

        self.redis_client.set_data_with_version = self.redis_client.register_script(
            LuaScripts.set_data_with_version())
        self.redis_client.get_latest_data = self.redis_client.register_script(
            LuaScripts.get_latest_data())
        self.redis_client.set_download_in_progress = self.redis_client.register_script(
            LuaScripts.set_download_in_progress())

    def validate_download_data(self, data):

        if hasattr(self, 'validate_' + self.slug) and callable(getattr(self, 'validate_' + self.slug)):
            valid, error_message = getattr(self, 'validate_' + self.slug)(data)
            if not valid:
                self.logger.error(f'Response from url did not validate', data={'error': error_message, **self.log_data})
                return
        return data

    def parse_download_data(self, data):

        if not self.version_subkey is None:
            if not isinstance(data, dict) or not self.version_subkey in data:
                self.logger.error(f'Response from url should be a JSON object containing the key {self.version_subkey}',
                                  data=self.log_data)
                return None, None
            version = str(data[self.version_subkey])
        else:
            # round this bigtime, otherwise multiple instances will cascade download files
            version = str(self.current_timestamp // 1814400 * 1814400)

        if not self.download_subkey is None:
            if not isinstance(data, dict) or not self.download_subkey in data:
                self.logger.error(f'Response from url should be a JSON object containing the key {self.download_subkey}',
                                  data=self.log_data)
                return None, None
            data = data[self.download_subkey]

        return version, data

    def download_in_progress(self):

        if not self.download_try_time is None and \
                self.download_try_time > self.current_time - datetime.timedelta(seconds=self.download_in_progress_ttl):
            return True
        self.download_try_time = self.current_time

        try:
            t = self.redis_client.set_download_in_progress(
                keys=[self.download_in_progress_key],
                args=[self.download_in_progress_ttl]
            )
            if t == b'0':
                return True
            return False
        except (redis.exceptions.TimeoutError, redis.exceptions.ConnectionError):
            return False

    def download(self):

        if self.download_in_progress():
            if self.verbose:
                print(f'{self.slug} - Download already in progress')
            return

        try:
            if self.verbose:
                print(f'{self.slug} - Downloading...')
            r = self.session.get(url=self.url, **self.download_options, timeout=self.download_timeout)
            r.raise_for_status()
            self.logger.info('Downloaded data from url',
                             data=self.log_data)
        except requests.exceptions.RequestException as e:
            self.logger.error('Failed to get data from url', data={'error': str(e), **self.log_data})
            return

        try:
            data = r.json()
        except json.JSONDecodeError:
            self.logger.error('Response from url was no valid json', data=self.log_data)
            return

        version, data = self.parse_download_data(data=data)
        if data is None:
            if self.verbose:
                print(f'{self.slug} - Downloaded data invalid')
            return

        data = self.validate_download_data(data=data)
        if data is None:
            if self.verbose:
                print(f'{self.slug} - Downloaded data did not validate')
            return

        download_time = self.current_time
        try:
            self.write_data_to_redis(data=data, version=version, download_time=download_time)
        except (redis.exceptions.TimeoutError, redis.exceptions.ConnectionError):
            pass
        self.write_data_to_file(data=data, version=version, download_time=download_time)

        self.load_content(data=data,
                          version=version,
                          download_time=download_time,
                          source='download')

    def write_data_to_file(self, data, version, download_time):

        data = self.build_data_object(content=data, version=version, download_time=download_time)
        file_path = self.build_file_path(version=version)
        with open(file_path, 'w') as f:
            json.dump(data, f)
        self.constrain_file_count()

    def constrain_file_count(self):

        latest_files = [fn for fn in [self.parse_filename(filename) for filename in os.listdir(self.data_dir)
                        if filename.endswith('.json') and filename.startswith(self.slug)
                        and not filename.endswith('fallback.json')] if not fn is None]
        if len(latest_files) > self.max_file_count:
            for file in sorted(latest_files, key=lambda x: int(x['version']))[:-3]:
                os.remove(file['file_path'])

    def parse_filename(self, filename):

        parts = filename.split('.')
        if len(parts) < 3:
            return None
        return {
            'slug': '.'.join(parts[:-2]),
            'version': parts[-2],
            'file_path': os.path.join(self.data_dir, filename)
        }


    def build_file_path(self, version):

        file_name = f'{self.slug}.{version}.json'
        return os.path.join(self.data_dir, file_name)

    def build_data_object(self, content, version, download_time):

        return {
            'content': content,
            'version': version,
            'timestamp': download_time.timestamp()
        }

    def write_data_to_redis(self, data, version, download_time):

        if self.verbose:
            print(f'{self.slug} - Writing data to redis')
        data = self.build_data_object(content=data, version=version, download_time=download_time)
        result = self.redis_client.set_data_with_version(
            keys=[
                self.current_version_key,
                self.data_key(version=version)
            ],
            args=[
                version,
                json.dumps(data),
                self.data_key_prefix,
                self.data_ttl
            ]
        )
        if int(result[0]) == 1:
            if self.verbose:
                print(f'{self.slug} - Wrote data to redis successful')
            return

        data = result[1]
        if self.verbose:
            print(f'{self.slug} - Version is cache is newer')
        if not data is None:
            cache_version, download_time, content = self.parse_cache_data(data=data)
            if self.is_larger_than_loaded_version(cache_version):
                self.load_content(data=content,
                                  version=cache_version,
                                  download_time=download_time,
                                  source='cache')

    @property
    def download_in_progress_key(self):
        return f'settings:{self.domain}:{self.slug}:download_in_progress'

    @property
    def current_version(self):
        v = self.redis_client.get(self.current_version_key)
        if v is None:
            return None
        return v.decode()

    @property
    def current_version_key(self):
        return f'settings:{self.domain}:{self.slug}:version'

    def data_key(self, version):
        return f'{self.data_key_prefix}{version}'

    @property
    def data_key_prefix(self):
        return f'settings:{self.domain}:{self.slug}:data:'

    @property
    def session(self):
        if not hasattr(self, '_session'):
            self._session = Session()

            retry_strategy = Retry(
                total=3,
                backoff_factor=1,
                status_forcelist=[429, 500, 502, 503, 504],
                allowed_methods=['GET', 'POST'],
                respect_retry_after_header=True
            )
            self._session.mount('https://', HTTPAdapter(max_retries=retry_strategy))

        return self._session

    @property
    def current_time(self):
        return pytz.utc.localize(datetime.datetime.utcnow())

    @property
    def current_timestamp(self):
        return int(self.current_time.timestamp())

    @property
    def age(self):
        if self.download_time is None:
            return None
        return self.current_time - self.download_time

    @property
    def too_old(self):
        if self.age is None:
            return False
        return self.age > self.max_age - datetime.timedelta(seconds=self.data_margin_seconds)

    @property
    def is_fallback(self):
        return self.version == self.fallback_version

    def _validate_string_only_array(self, raw):
        if not isinstance(raw, list):
            return False, 'Should be a list'
        if not all(isinstance(r, str) for r in raw):
            return False, 'All entries should be strings'
        if not all(not r.strip() == '' for r in raw):
            return False, 'Empty string not allowed as entry'
        return True, None

    def _validate_dict(self, raw, type=str):
        if not isinstance(raw, dict):
            return False, 'Should be a dictionary'
        if not all(isinstance(k, str) for k in raw):
            return False, 'All keys should be strings'
        if not all(isinstance(v, type) for v in raw.values()):
            return False, f'All values should be {type.__name__}'
        return True, None

    def health(self):

        if self.content is None:
            self.get(version=self.debug_version)

        healthy = False

        if self.is_fallback and not self.fallback_allowed:
            reason = 'Fallback in use'
        elif not self.download_time is None and self.current_time > self.download_time + self.max_age:
            reason = 'Settings too old'
        else:
            healthy = True
            reason = None

        d = {
            'healthy': healthy,
            'slug': self.slug,
            'load_time': self.load_time.isoformat(),
            'download_time': self.download_time.isoformat(),
            'age': str(self.age),
            'max_age': str(self.max_age),
            'allow_fallback': self.fallback_allowed,
            'fallback': self.is_fallback,
            'source': self.source,
            'version': self.version
        }
        if not healthy:
            d.update({'reason': reason})
        return d

    @property
    def log_data(self):

        return {'slug': self.slug,
                'instance_id': self.instance_id,
                'url': self.url}




from python_grains.utils.utils import RedisClientWithRetry
from python_grains.utils.utils import chunker
from python_grains.utils.utils import search_obj_from_array
from python_grains.utils.utils import Timer


import os

from python_grains.profiles.profiles import Profile
from python_grains.profiles.profile_properties import (PROFILE_VALUE_CREATED_POSTFIX,
                                                       PROFILE_VALUE_UPDATED_POSTFIX,
                                                       PROFILE_VALUE_EXPIRE_AT_POSTFIX,
                                                       PROFILE_VALUE_POSTFIX)
from python_grains.profiles.lua import LuaScripts

def attach_profile_lua_scripts(redis_client,
                               user_ttl,
                               user_data_ttl_margin,
                               profile_ttl,
                               profile_ttl_margin,
                               new_user_detection_ttl):

    if not hasattr(redis_client, 'get_user_profile'):

        redis_client.get_user_profile = redis_client.register_script(
            LuaScripts.get_user_profile(
                value_postfix=PROFILE_VALUE_POSTFIX,
                updated_postfix=PROFILE_VALUE_UPDATED_POSTFIX,
                created_postfix=PROFILE_VALUE_CREATED_POSTFIX,
                expire_postfix=PROFILE_VALUE_EXPIRE_AT_POSTFIX))

    if not hasattr(redis_client, 'update_profile') and \
            not any(x is None for x in (user_ttl, user_data_ttl_margin)):

        redis_client.update_profile = redis_client.register_script(
            LuaScripts.update_profile(
                user_ttl=user_ttl,
                user_ttl_margin=user_data_ttl_margin,
                profile_ttl=profile_ttl,
                profile_ttl_margin=profile_ttl_margin,
                value_postfix=PROFILE_VALUE_POSTFIX,
                updated_postfix=PROFILE_VALUE_UPDATED_POSTFIX,
                created_postfix=PROFILE_VALUE_CREATED_POSTFIX,
                expire_postfix=PROFILE_VALUE_EXPIRE_AT_POSTFIX,
                new_user_detection_ttl=new_user_detection_ttl))

    if not hasattr(redis_client, 'get_all_user_pointers') and \
            not any(x is None for x in (user_ttl, user_data_ttl_margin)):

        redis_client.get_all_user_pointers = redis_client.register_script(
            LuaScripts.get_all_user_pointers(
                user_ttl=user_ttl,
                user_ttl_margin=user_data_ttl_margin
            ))

    if not hasattr(redis_client, 'set_profile_if_new_user') and \
            not any(x is None for x in (user_ttl, user_data_ttl_margin, new_user_detection_ttl)):

        redis_client.set_profile_if_new_user = redis_client.register_script(
            LuaScripts.set_profile_if_new_user(
                user_ttl=user_ttl,
                user_ttl_margin=user_data_ttl_margin,
                profile_ttl=profile_ttl,
                profile_ttl_margin=profile_ttl_margin,
                new_user_detection_ttl=new_user_detection_ttl
            ))

    if not hasattr(redis_client, 'completely_remove_from_cache'):

        redis_client.completely_remove_from_cache = redis_client.register_script(
            LuaScripts.completely_remove_from_cache())

    if not hasattr(redis_client, 'get_multiple_stitches'):

        redis_client.get_multiple_stitches = redis_client.register_script(
            LuaScripts.get_multiple_stitches())

    if not hasattr(redis_client, 'get_stitched_user_profile'):

        redis_client.get_stitched_user_profile = redis_client.register_script(
            LuaScripts.get_stitched_user_profile(
                value_postfix=PROFILE_VALUE_POSTFIX,
                updated_postfix=PROFILE_VALUE_UPDATED_POSTFIX,
                created_postfix=PROFILE_VALUE_CREATED_POSTFIX,
                expire_postfix=PROFILE_VALUE_EXPIRE_AT_POSTFIX
            ))

    if not hasattr(redis_client, 'switch_profiles') and \
            not any(x is None for x in (user_ttl, user_data_ttl_margin)):

        redis_client.switch_profiles = redis_client.register_script(
            LuaScripts.switch_profiles(
                user_ttl=user_ttl,
                user_ttl_margin=user_data_ttl_margin
            ))

    if not hasattr(redis_client, 'finalize_profile'):

        redis_client.finalize_profile = redis_client.register_script(
            LuaScripts.finalize_profile(local=bool(os.getenv('DJANGO_DEVELOPMENT'))))

    if not hasattr(redis_client, 'delete_in_cache'):

        redis_client.delete_in_cache = redis_client.register_script(
            LuaScripts.delete_in_cache())

    if not hasattr(redis_client, 'get_all_user_pointers_if_main_profile'):

        redis_client.get_all_user_pointers_if_main_profile = redis_client.register_script(
            LuaScripts.get_all_user_pointers_if_main_profile())

    return redis_client

def getProfileClass(
        domain,
        fast_clean_threshold,
        redis_client,
        django_profile_model,
        django_pointer_model,
        django_connection_tuple,  # should contain (connection, connections) from django.db
        django_count_func,
        profile_ttl,
        profile_ttl_margin,
        user_ttl=None,
        user_data_ttl_margin=None,
        clean_queue_ttl=None,
        new_user_detection_ttl=None):

    cls = Profile
    cls.domain = domain
    cls.fast_clean_threshold = fast_clean_threshold
    cls.redis_client = attach_profile_lua_scripts(redis_client=redis_client,
                                                  user_ttl=user_ttl,
                                                  user_data_ttl_margin=user_data_ttl_margin,
                                                  profile_ttl=profile_ttl,
                                                  profile_ttl_margin=profile_ttl_margin,
                                                  new_user_detection_ttl=new_user_detection_ttl)
    cls.django_profile_model = django_profile_model
    cls.django_pointer_model = django_pointer_model
    cls.django_connection_tuple = django_connection_tuple
    cls.django_count_func = django_count_func
    cls.profile_ttl = profile_ttl
    cls.profile_ttl_margin = profile_ttl_margin
    cls.user_ttl = user_ttl
    cls.user_data_ttl_margin = user_data_ttl_margin
    cls.clean_queue_ttl = clean_queue_ttl

    return cls


import datetime
import pytz
import json

PROFILE_VALUE_POSTFIX = '_v'
PROFILE_VALUE_UPDATED_POSTFIX = '_u'
PROFILE_VALUE_CREATED_POSTFIX = '_c'
PROFILE_VALUE_EXPIRE_AT_POSTFIX = '_e'

class ProfilePropertyError(Exception): pass


class ProfileProperty(object):

    type = None
    prefix = None

    def __init__(self,
                 name,
                 value=None,
                 updated=None,
                 created=None,
                 ttl=None,
                 expire_at=None):

        self.name = name
        self.value = self.validate(value) if not value is None else value

        now = pytz.utc.localize(datetime.datetime.utcnow())
        self.updated = updated or now
        self.created = created or now

        if ttl:
            self.expire_at = now + datetime.timedelta(seconds=ttl)
        else:
            self.expire_at = expire_at or None

    def __repr__(self):

        return str(self.__class__.__name__) + '(`{}`:{})'.format(self.name, self.value)

    def validate(self,
                 value):

        if self.type:
            try:
                value = self.type(value)
            except (ValueError, TypeError):
                raise ProfilePropertyError(f'Value must be of type (or coercible to): {self.type.__name__}')

        return value

    def __add__(self,
                other):

        if isinstance(other, int) and other == 0:
            return self

        assert self.__class__ == other.__class__, 'Cannot add {} to {}'.format(self.__class__.__name__,
                                                                               other.__class__.__name__)
        assert self.name == other.name, 'Can only add properties with the same name.'
        assert isinstance(self.value, type(other.value)), 'Cannot add {} to {}'.format(type(self.value), type(other.value))

        return self.combine(other)

    def __radd__(self,
                 other):

        return self.__add__(other)

    def combine(self,
                other):

        new_value = self.value + other.value
        new_updated = max([self.updated, other.updated])
        new_created = min([self.created, other.created])
        new_expire_at = self.combine_expire(other)

        return self.__class__(self.name, new_value, updated=new_updated, created=new_created, expire_at=new_expire_at)

    def combine_expire(self,
                       other):

        if self.expire_at is None and other.expire_at is None:
            return None
        if self.expire_at is None:
            return other.expire_at
        if other.expire_at is None:
            return self.expire_at
        if self.updated > other.updated:
            return self.expire_at
        if self.updated < other.updated:
            return other.updated
        return max([self.expire_at, other.expire_at])

    def copy_off(self, obj, expire_at):

        return self.__class__(obj.name, obj.value, updated=obj.updated, created=obj.created, expire_at=expire_at)

    @classmethod
    def from_redis_data(cls,
                        name,
                        value=None,
                        updated=None,
                        created=None,
                        ttl=None,
                        expire_at=None):

        return cls(name=name,
                   value=cls.value_from_redis(value),
                   updated=updated,
                   created=created,
                   ttl=ttl,
                   expire_at=expire_at)

    @classmethod
    def value_from_redis(cls, value):
        return value

    @classmethod
    def value_to_redis(cls, value):

        return value

    @property
    def expired(self):

        if self.expire_at:
            return self.expire_at < pytz.utc.localize(datetime.datetime.utcnow())
        return False

    @property
    def output_value(self):
        return self.value

    @property
    def redis_keys(self):

        s = self.prefix + self.name
        d = {'key': s}
        if not self.value is None:
            d.update({'value': self.value_to_redis(self.value)})
        if self.expire_at:
            d.update({'expire_at': self.expire_at.timestamp()})
        if self.extra_redis_args:
            d.update(self.extra_redis_args)
        return d

    @property
    def extra_redis_args(self):
        return {}

    @property
    def flat_redis_keys(self):

        s = self.prefix + self.name
        d = {
            s + PROFILE_VALUE_POSTFIX: self.value_to_redis(self.value),
            s + PROFILE_VALUE_UPDATED_POSTFIX: self.updated.timestamp(),
            s + PROFILE_VALUE_CREATED_POSTFIX: self.created.timestamp()
        }
        if self.expire_at:
            d.update({s + PROFILE_VALUE_EXPIRE_AT_POSTFIX: self.expire_at.timestamp()})
        return d

    @property
    def short_repr(self):

        return {self.name: self.output_value}

    @property
    def dict_repr(self):

        d = {
            'name': self.name,
            'value': self.output_value,
            'updated': self.updated.isoformat(),
            'created': self.created.isoformat()
        }

        if self.expire_at:
            d['expire_at'] = self.expire_at.isoformat()

        return d

class ProfileString(ProfileProperty):

    type = str


class ProfileFloat(ProfileProperty):

    type = float


class ProfileBoolean(ProfileProperty):

    type = bool

    @classmethod
    def value_from_redis(cls, value):

        return value in ('1', b'1', 'True', b'True', 'true', b'true', True, 1)

    @classmethod
    def value_to_redis(cls, value):

        return int(value)


class OR(object):

    def combine(self, other):
        new_value = self.value or other.value
        new_updated = max([self.updated, other.updated])
        new_created = min([self.created, other.created])
        new_expire_at = self.combine_expire(other)

        return self.__class__(self.name, new_value, updated=new_updated, created=new_created, expire_at=new_expire_at)


class AND(object):

    def combine(self, other):
        new_value = self.value and other.value
        new_updated = max([self.updated, other.updated])
        new_created = min([self.created, other.created])
        new_expire_at = self.combine_expire(other)

        return self.__class__(self.name, new_value, updated=new_updated, created=new_created, expire_at=new_expire_at)


class LastUpdated(object):

    def combine(self, other):
        expire_at = self.combine_expire(other)

        if self.updated < other.updated:
            return self.copy_off(other, expire_at=expire_at)
        return self.copy_off(self, expire_at=expire_at)


class LastCreated(object):

    def combine(self, other):
        expire_at = self.combine_expire(other)

        if self.created < other.created:
            return self.copy_off(other, expire_at=expire_at)
        return self.copy_off(self, expire_at=expire_at)


class FirstUpdated(object):

    def combine(self, other):
        expire_at = self.combine_expire(other)

        if self.updated > other.updated:
            return self.copy_off(other, expire_at=expire_at)
        return self.copy_off(self, expire_at=expire_at)


class FirstCreated(object):

    def combine(self, other):
        expire_at = self.combine_expire(other)

        if self.created > other.created:
            return self.copy_off(other, expire_at=expire_at)
        return self.copy_off(self, expire_at=expire_at)


class SmallestValue(object):

    def combine(self, other):
        expire_at = self.combine_expire(other)

        if self.value > other.value:
            return self.copy_off(other, expire_at=expire_at)
        return self.copy_off(self, expire_at=expire_at)


class LargestValue(object):

    def combine(self, other):

        expire_at = self.combine_expire(other)

        if self.value < other.value:
            return self.copy_off(other, expire_at=expire_at)
        return self.copy_off(self, expire_at=expire_at)


class CounterWithTimeWindow(ProfileProperty):

    window_step_size = datetime.timedelta(days=1)
    count_from = datetime.datetime(2021, 1, 1, 0, 0, 0)
    timezone = pytz.timezone('Europe/Amsterdam')

    @classmethod
    def threshold_date(cls):

        return pytz.utc.localize(datetime.datetime.now()).astimezone(cls.timezone) \
               - cls.window_size * cls.window_step_size

    @classmethod
    def threshold_day_counter(cls):

        return (cls.threshold_date().date() - cls.count_from.date()).days

    @classmethod
    def current_day_counter(cls):

        return (pytz.utc.localize(datetime.datetime.now()).astimezone(cls.timezone).date() - cls.count_from.date()).days

    @classmethod
    def _current_value(cls, value):

        return {ts: v for ts,v in value.items() if int(ts) > cls.threshold_day_counter()}

    @property
    def current_value(self):

        return self._current_value(value=self.value)

    @classmethod
    def value_from_redis(cls, value):
        value = value.decode() if isinstance(value, bytes) else value
        data = {int(ts): int(val) for ts, val in json.loads(value).items()}
        return cls._current_value(data)

    @classmethod
    def value_to_redis(cls, value):

        return json.dumps(cls._current_value(value))

    @property
    def output_value(self):

        return sum(self.current_value.values())

    def combine(self, other):

        new_updated = max([self.updated, other.updated])
        new_created = min([self.created, other.created])
        new_expire_at = self.combine_expire(other)

        new_value = self.current_value
        for ts, val in other.current_value.items():
            new_value[ts] = new_value.get(ts, 0) + val

        return self.__class__(self.name, new_value, updated=new_updated, created=new_created, expire_at=new_expire_at)

    @property
    def extra_redis_args(self):
        return {
            'thr': self.threshold_day_counter()}


class UsableProperty(object): pass

###### Use one of these: #######

class Counter(ProfileProperty, UsableProperty):
    type = int
    prefix = 'ctr_'
    py_key = 'counters'
    js_key = 'counters'

class CounterWithTimeWindow30(CounterWithTimeWindow, UsableProperty):
    type = dict
    prefix = 'tmctr30_'
    py_key = 'time_window30_counters'
    js_key = 'timeWindow30Counters'
    window_size = 30

class SummedFloat(ProfileFloat, UsableProperty):
    prefix = 'sumflt_'
    py_key = 'summed_floats'
    js_key = 'summedFloats'


class LastUpdatedFloat(LastUpdated, ProfileFloat, UsableProperty):
    prefix = 'luflt_'
    py_key = 'last_updated_floats'
    js_key = 'lastUpdatedFloats'


class FirstCreatedFloat(FirstCreated, ProfileFloat, UsableProperty):
    prefix = 'fcflt_'
    py_key = 'first_created_floats'
    js_key = 'firstCreatedFloats'


class SmallestFloat(SmallestValue, ProfileFloat, UsableProperty):
    prefix = 'sflt_'
    py_key = 'smallest_floats'
    js_key = 'smallestFloats'


class LargestFloat(LargestValue, ProfileFloat, UsableProperty):
    prefix = 'lflt_'
    py_key = 'largest_floats'
    js_key = 'largestFloats'


class LastUpdatedString(LastUpdated, ProfileString, UsableProperty):
    prefix = 'lustr_'
    py_key = 'last_updated_strings'
    js_key = 'lastUpdatedStrings'


class FirstCreatedString(FirstCreated, ProfileString, UsableProperty):
    prefix = 'fcstr_'
    py_key = 'first_created_strings'
    js_key = 'firstCreatedStrings'


class OrBoolean(OR, ProfileBoolean, UsableProperty):
    prefix = 'obln_'
    py_key = 'or_booleans'
    js_key = 'orBooleans'


class AndBoolean(AND, ProfileBoolean, UsableProperty):
    prefix = 'abln_'
    py_key = 'and_booleans'
    js_key = 'andBooleans'



import hashlib

DEFAULT_N_CHARS = 10  # should not exceed 32 (half the length of the hex representation of a sha256 hash)
DEFAULT_EXPERIMENTS_RANGE_SEED = 'wjBr2FiEdFUKMz523E7Ro3kuSc5v5Oxh'

def experiment_init(self):
    self.variants = [v() for v in self.variants]
    self.initialize()

def variant_init(self):
    self.initialize()

class Variant(object):

    def __init__(self):
        raise NotImplementedError('Class should be setup by using Variant.setup')

    def initialize(self):
        pass

    @classmethod
    def setup(cls, **kwargs):

        mandatory_kwargs = {
            'id',
            'name',
            'prop' }

        defaults = {
            'n_chars': DEFAULT_N_CHARS }

        missing = mandatory_kwargs - set(kwargs.keys())
        if len(missing) > 0:
            raise Exception(f'Missing configurations: {", ".join(missing)}')

        for prop, value in defaults.items():
            if not prop in kwargs:
                kwargs[prop] = value

        if not '__init__' in kwargs:
            kwargs['__init__'] = variant_init

        return type(cls.__name__ + '_' + kwargs['id'], (cls,), kwargs)

    def user_in_variant(self,
                        user_id):

        p = int(hashlib.sha256((str(user_id) + self.seed).encode('utf-8')).hexdigest()[:self.n_chars], 16) / 16 ** self.n_chars
        return p <= self.cum_prob

    @property
    def data(self):

        return {
            'id': self.id,
            'name': self.name
        }

class Experiment(object):

    def __init__(self):
        raise NotImplementedError('Class should be setup by using Experiment.setup')

    def __repr__(self):
        return str('<Experiment: ' + self.name + '>')

    def initialize(self):
        pass

    @classmethod
    def setup(cls, **kwargs):

        mandatory_kwargs = {
            'id',
            'name',
            'seed'}

        defaults = {
            'use_range': False,
            'range_start': None,
            'range_end': None,
            'hostnames': None,
            'origin_type': 'all',
            'n_chars': DEFAULT_N_CHARS,
            'variants': []
        }

        missing = mandatory_kwargs - set(kwargs.keys())
        if len(missing) > 0:
            raise Exception(f'Missing configurations: {", ".join(missing)}')

        if 'use_range' in kwargs and kwargs['use_range'] and not 'range_seed' in kwargs:
            raise Exception(f'Must provide range_seed if "use_range==True"')

        if 'variants' in kwargs:
            sort_key = kwargs.get('variant_sort_key') or 'id'
            kwargs['variants'] = cls.prepare_variants(kwargs['variants'], seed=kwargs['seed'], sort_key=sort_key)

        for prop, value in defaults.items():
            if not prop in kwargs:
                kwargs[prop] = value

        kwargs['__init__'] = experiment_init

        return type(cls.__name__ + '_' + kwargs['id'], (cls,), kwargs)

    def set_variants(self,
                     variants,
                     sort_key='id'):

        self.variants = self.prepare_variants(variants, self.seed, sort_key=sort_key)

    @staticmethod
    def prepare_variants(variants, seed, sort_key='id'):

        assert all(isinstance(var, Variant) or issubclass(var, Variant) for var in variants)

        variants = sorted(variants, key=lambda v: getattr(v, sort_key))
        for variant in variants:
            variant.seed = seed

        total_prop = sum(v.prop for v in variants)
        cum_prob = 0
        for variant in variants:
            cum_prob += variant.prop / total_prop
            variant.cum_prob = cum_prob

        return variants

    def origin_type_allowed(self,
                            origin_type):

        return self.origin_type == 'all' or origin_type.lower() == self.origin_type

    def hostname_match(self,
                       hostname,
                       match_string):

        if match_string.startswith('.'):
            return hostname.lower().endswith(match_string[1:])
        else:
            return hostname.lower() == match_string

    def hostname_allowed(self,
                         hostname):

        if not self.hostnames:
            return True
        return any(self.hostname_match(hostname, m) for m in self.hostnames)

    def user_in_experiment(self,
                           user_id,
                           origin_type,
                           hostname):

        if not self.origin_type_allowed(origin_type=origin_type):
            return False
        if not self.hostname_allowed(hostname=hostname):
            return False

        if self.use_range:
            p = int(
                hashlib.sha256((str(user_id) + self.range_seed).encode('utf-8')).hexdigest()[-self.n_chars:],
                16) / 16 ** self.n_chars
            if not self.range_start < p <= self.range_end:
                return False

        return True

    def get_variant_for_user(self,
                             user_id):

        for variant in self.variants[:-1]:
            if variant.user_in_variant(user_id):
                return variant
        return self.variants[-1]

    def data_for_user(self,
                      user_id,
                      verbose=False):
        variant = self.get_variant_for_user(user_id=user_id)
        variant_data = variant.data
        variant_id = variant_data['id']

        if verbose:

            return {
                'id': self.id,
                'name': self.name,
                'expVarId': self.id + ':' + variant_id,
                'variant': variant.data}

        return {
            'id': self.id,
            'expVarId': self.id + ':' + variant_id,
            'variant': {'id': variant.data['id']}}


class LuaScripts(object):

    @classmethod
    def create_processing_queue(cls, data_idx_key, hash_event_prefix, legacy=False):

        '''
        KEYS[1]: main_queue_key
        KEYS[2]: processing_queue_key
        KEYS[3]: register processing queue key
        ARGV[1]: min_ts
        ARGV[2]: max_ts
        ARGV[3]: ttl processing queue
        ARGV[4]: ts start processing queue
        ARGV[4]: max elements
        '''

        if legacy:

            return '''
                local elements = redis.call('ZRANGEBYSCORE', KEYS[1], ARGV[1], ARGV[2], 'LIMIT', 0, ARGV[5])
                local payload = {{}}
                
                for idx, el in pairs(elements) do
                    local p = cjson.decode(el)
                    table.insert(payload, '{hash_event_prefix}:' .. p['{data_idx_key}'])
                    table.insert(payload, el)
                end
                if #payload > 0 then
                    redis.call('HMSET', KEYS[2], unpack(payload))
                    redis.call('ZREM', KEYS[1], unpack(elements))
                    redis.call('EXPIRE', KEYS[2], ARGV[3])
                    redis.call('ZADD', KEYS[3], ARGV[4], KEYS[2])
                    redis.call('EXPIRE', KEYS[3], ARGV[3])
                end
                return elements'''.format(data_idx_key=data_idx_key, hash_event_prefix=hash_event_prefix)

        else:

            return '''
                local elements = redis.call('ZRANGE', KEYS[1], ARGV[1], ARGV[2], 'BYSCORE', 'LIMIT', 0, ARGV[5])
                local payload = {{}}
                
                for idx, el in pairs(elements) do
                    local p = cjson.decode(el)
                    table.insert(payload, '{hash_event_prefix}:' .. p['{data_idx_key}'])
                    table.insert(payload, el)
                end
                if #payload > 0 then
                    redis.call('HSET', KEYS[2], unpack(payload))
                    redis.call('ZREM', KEYS[1], unpack(elements))
                    redis.call('EXPIRE', KEYS[2], ARGV[3])
                    redis.call('ZADD', KEYS[3], ARGV[4], KEYS[2])
                    redis.call('EXPIRE', KEYS[3], ARGV[3])
                end
                return elements'''.format(data_idx_key=data_idx_key, hash_event_prefix=hash_event_prefix)

    @classmethod
    def delete_queue(cls):

        '''
        KEYS[1]: main queue key
        KEYS[2]: register processing queue key
        '''

        return '''
            redis.call('DEL', KEYS[1])
            local rogue_queue_keys = redis.call('ZRANGE', KEYS[2], 0, -1)
            if #rogue_queue_keys > 0 then
                redis.call('DEL', unpack(rogue_queue_keys))
            end
            redis.call('DEL', KEYS[2])
       '''




import datetime
import pytz
import uuid
import json
import time
import hashlib
from redis.exceptions import TimeoutError, ConnectionError
from contextlib import ExitStack
from inspect import signature

from python_grains.utils import chunker
from python_grains.locks import Lock, LockError
from python_grains.queues.lua import LuaScripts

DATA_IDX_KEY = 'i'
DATA_TIME_KEY = 't'
DATA_DATA_KEY = 'd'
DATA_MAX_RETRY_KEY = 'm'
HASH_EVENT_PREFIX = 'e'

DEFAULT_QUEUE_TTL = 24 * 60 * 60
DEFAULT_PROCESSING_QUEUE_TTL = 60 * 60
DEFAULT_LOCK_TTL = 10
DEFAULT_MAX_RETRY = 1

class QueueError(Exception): pass

class Event(object):

    data_idx_key = DATA_IDX_KEY
    data_time_key = DATA_TIME_KEY
    data_data_key = DATA_DATA_KEY
    data_max_retry_key = DATA_MAX_RETRY_KEY
    default_lock_ttl = DEFAULT_LOCK_TTL

    def __init__(self,
                 data,
                 redis_client,
                 timestamp=None,
                 idx=None,
                 lock_ttl=None,
                 processing_queue_key=None,
                 max_retry=DEFAULT_MAX_RETRY):

        self.data = data
        self.datetime = self.datetime_from_utctimestamp(timestamp) or self.utcnow()
        self.idx = idx or self.new_idx()
        self.redis_client = redis_client
        self.lock_ttl = lock_ttl or self.default_lock_ttl
        self.processing_queue_key = processing_queue_key
        self.max_retry = max_retry

        self.processed = False
        self.n_try = 0
        self.success = False
        self.result = None

    @staticmethod
    def datetime_from_utctimestamp(timestamp):

        if timestamp is None:
            return None
        return pytz.utc.localize(datetime.datetime.utcfromtimestamp(timestamp))

    @staticmethod
    def utcnow():

        return pytz.utc.localize(datetime.datetime.utcnow())

    @staticmethod
    def new_idx():

        return str(uuid.uuid4())

    def lock(self):

        return Lock(idx=self.lock_idx,
                    conn=self.redis_client,
                    ttl=self.lock_ttl * 1000)

    def delete(self, pipe=None):
        if pipe:
            pipe.hdel(self.processing_queue_key,
                      self.try_count_key,
                      self.result_key)
        else:
            self.redis_client.hdel(self.processing_queue_key,
                                   self.try_count_key,
                                   self.result_key)

    def store_result(self, result, pipe=None):

        assert isinstance(result, str), 'result should be a string'

        if pipe:
            pipe.hset(self.processing_queue_key,
                                   self.result_key,
                                   result)
        else:
            self.redis_client.hset(self.processing_queue_key,
                                   self.result_key,
                                   result)


    def reset_try_count(self):

        self.redis_client.hdel(self.processing_queue_key,
                               self.try_count_key)

    def incr_try_count(self, pipe=None):

        if self.processing_queue_key is None:
            raise QueueError('No processing queue key given')

        if pipe:
            pipe.hincrby(self.processing_queue_key,
                                      self.try_count_key,
                                      amount=1)
        else:
            self.redis_client.hincrby(self.processing_queue_key,
                                      self.try_count_key,
                                      amount=1)

    @property
    def max_tries_reached(self):

        return self.try_count >= self.max_retry + 1

    @property
    def prev_result(self):
        result = self.redis_client.hget(self.processing_queue_key,
                                        self.result_key)
        return result.decode() if isinstance(result, bytes) else result

    @property
    def try_count(self):

        if self.processing_queue_key is None:
            raise QueueError('No processing queue key given')

        count = self.redis_client.hget(self.processing_queue_key,
                                       self.try_count_key)
        if count is None:
            return 0

        return int(count)

    @classmethod
    def from_redis_string(cls,
                          raw_string,
                          redis_client,
                          processing_queue_key=None):

        raw_string = raw_string.decode() if isinstance(raw_string, bytes) else raw_string
        data = json.loads(raw_string)
        return cls(data=data[cls.data_data_key],
                   timestamp=data[cls.data_time_key],
                   idx=data[cls.data_idx_key],
                   max_retry=data[cls.data_max_retry_key],
                   redis_client=redis_client,
                   processing_queue_key=processing_queue_key)

    @property
    def score(self):

        return self.datetime.timestamp()

    @property
    def event_data(self):

        return {
            self.data_idx_key: self.idx,
            self.data_time_key: self.score,
            self.data_data_key: self.data,
            self.data_max_retry_key: self.max_retry
        }

    @property
    def datastring(self):

        return json.dumps(self.event_data)

    @property
    def mapping(self):

        return {self.datastring: self.score}

    @property
    def lock_idx(self):

        return f'ev:{self.idx}'

    @property
    def redis_version(self):

        if not hasattr(self, '_redis_version'):
            self._redis_version = tuple(int(i) for i in self.redis_client.info()['redis_version'].split('.'))

        return self._redis_version

    @property
    def try_count_key(self):

        return f'rt:{self.idx}'

    @property
    def result_key(self):

        return f'rs:{self.idx}'

    @property
    def _ident(self):
        conn_kwargs = self.redis_client.get_connection_kwargs()
        return f'{conn_kwargs["host"]}:{conn_kwargs["port"]}:{conn_kwargs["db"]}:{self.idx}'

    @property
    def guid(self):
        return hashlib.sha256(self._ident.encode('utf-8')).hexdigest()

class ProcessingQueue(object):

    key_prefix = 'qp'
    default_ttl = DEFAULT_PROCESSING_QUEUE_TTL
    lock_ttl = DEFAULT_LOCK_TTL

    def __init__(self,
                 parent_key,
                 register_key,
                 main_queue_ttl,
                 redis_client,
                 idx=None,
                 ttl=None):

        self.parent_key = parent_key
        self.register_key = register_key
        self.main_queue_ttl = main_queue_ttl
        self.ttl = ttl or self.default_ttl
        self.idx = idx or self.new_idx()
        self.events = None
        self.redis_client = redis_client
        self.add_create_processing_queue_lua()

    def add_create_processing_queue_lua(self):

        if not hasattr(self.redis_client, 'create_processing_queue'):
            self.redis_client.create_processing_queue = self.redis_client.register_script(
                LuaScripts.create_processing_queue(data_idx_key=DATA_IDX_KEY,
                                                   hash_event_prefix=HASH_EVENT_PREFIX,
                                                   legacy=self.redis_version[0] < 4))

    def new_idx(self):

        return str(uuid.uuid4())

    def setup(self,
              max_elements,
              min_ts,
              max_ts):

        result = self.redis_client.create_processing_queue(
            keys=[
                self.parent_key,
                self.queue_key,
                self.register_key
            ],
            args=[
                min_ts,
                max_ts,
                self.main_queue_ttl,
                pytz.utc.localize(datetime.datetime.utcnow()).timestamp(),
                max_elements
            ]
        )
        self.events = [Event.from_redis_string(raw_string=raw_string,
                                               redis_client=self.redis_client,
                                               processing_queue_key=self.queue_key)
                       for raw_string in result]

    def lock(self):

        return Lock(idx=self.lock_idx,
                    conn=self.redis_client,
                    ttl=self.lock_ttl * 1000)

    def delete(self):

        pipe = self.redis_client.pipeline()
        pipe.delete(self.queue_key)
        pipe.zrem(self.register_key, self.queue_key)
        pipe.execute()

    @classmethod
    def get(cls,
            parent_key,
            register_key,
            main_queue_ttl,
            redis_client,
            max_elements,
            min_ts,
            max_ts,
            ttl=None):

        queue = cls(parent_key=parent_key,
                    register_key=register_key,
                    main_queue_ttl=main_queue_ttl,
                    ttl=ttl,
                    redis_client=redis_client)

        queue.setup(min_ts=min_ts,
                    max_ts=max_ts,
                    max_elements=max_elements)

        return queue

    def result_object(self, event_id, success, result, will_retry):

        return {
            'id': event_id,
            'success': success,
            'result': result,
            'retry': will_retry}

    def process(self,
                func):

        n_par = len(signature(func).parameters)
        if n_par == 0 or n_par > 2:
            raise QueueError('processing function should take 1 or 2 arguments')

        results = []

        for event in self.events:
            try:
                with event.lock():
                    if n_par == 1:
                        success, result = func(event.data)
                    else:
                        success, result = func(event.data, event.prev_result)
                    if success:
                        event.delete()
                        will_retry = False
                    else:
                        if not event.max_tries_reached:
                            event.store_result(result=result)
                            event.incr_try_count()
                            will_retry = True
                        else:
                            event.delete()
                            will_retry = False
                    results.append(self.result_object(event_id=event.idx,
                                                      success=success,
                                                      result=result,
                                                      will_retry=will_retry))
            except (LockError, TimeoutError, ConnectionError) as e:
                will_retry = True
                if isinstance(e, LockError):
                    success = None
                    result = 'Lock failed'
                else:
                    success = False
                    result = str(e)
                    event.incr_try_count()
                    if not event.max_tries_reached:
                        event.store_result(result=result)
                        event.incr_try_count()
                    else:
                        event.delete()
                        will_retry = False

                results.append(self.result_object(event_id=event.idx, success=success, result=result, will_retry=will_retry))
                continue

        if all(result['success'] or not result['retry'] for result in results):
            self.delete()

        return results

    def batch_process(self,
                      func,
                      batch_size):

        results = []
        for chunk in chunker(self.events, batch_size):
            try:
                locks = [event.lock() for event in chunk]
                with ExitStack() as stack:
                    for lock in locks:
                        stack.enter_context(lock)
                    _results = func([(event.idx, event.data, event.prev_result) for event in chunk])
                    _event_dict = {event.idx: event for event in chunk}
                    pipe = self.redis_client.pipeline()
                    for r in _results:
                        event_id, success, result = r
                        event = _event_dict[event_id]
                        if success:
                            event.delete(pipe=pipe)
                            will_retry = False
                        else:
                            if not event.max_tries_reached:
                                event.store_result(result=result, pipe=pipe)
                                event.incr_try_count(pipe=pipe)
                                will_retry = True
                            else:
                                event.delete(pipe=pipe)
                                will_retry = False
                        results.append(self.result_object(event_id=event.idx,
                                                          success=success,
                                                          result=result,
                                                          will_retry=will_retry))
                    pipe.execute()


            except (LockError, ConnectionError, TimeoutError) as e:
                will_retry = True
                if isinstance(e, LockError):
                    success = None
                    result = 'Lock failed'
                else:
                    success = False
                    result = str(e)
                results.extend([self.result_object(event_id=event.idx,
                                                   success=success,
                                                   result=result,
                                                   will_retry=will_retry) for event in chunk])
                continue

            if all(result['success'] or not result['retry'] for result in results):
                self.delete()

        return results

    def setup_rogue(self):

        result = self.redis_client.hgetall(self.queue_key)
        if len(result) > 0:
            result_dict = {k.decode() if isinstance(k, bytes) else k: v for k, v in result.items()}
            event_dict = {k: v for k, v in result_dict.items() if k.startswith(HASH_EVENT_PREFIX + ':')}
            self.events = [Event.from_redis_string(raw_string=raw_string,
                                                   redis_client=self.redis_client,
                                                   processing_queue_key=self.queue_key)
                           for raw_string in event_dict.values()]
        else:
            self.events = []

        if len(self.events) > 0:
            self.redis_client.expire(self.queue_key, self.main_queue_ttl)

    @classmethod
    def get_rogue(cls,
                  parent_key,
                  register_key,
                  main_queue_ttl,
                  redis_client,
                  margin=120,
                  offset=0,
                  ttl=None):

        max_ts = pytz.utc.localize(datetime.datetime.utcnow()).timestamp() - margin
        queue_keys = redis_client.zrangebyscore(register_key,
                                                min='-inf',
                                                max=max_ts,
                                                start=offset,
                                                num=1)

        if len(queue_keys) == 0:
            return None

        queue_key = queue_keys[0]
        queue_key = queue_key.decode() if isinstance(queue_key, bytes) else queue_key
        queue_idx = queue_key.split(':')[-1]
        queue = cls(parent_key=parent_key,
                    register_key=register_key,
                    main_queue_ttl=main_queue_ttl,
                    ttl=ttl,
                    idx=queue_idx,
                    redis_client=redis_client)
        queue.setup_rogue()
        return queue

    @property
    def redis_version(self):

        if not hasattr(self, '_redis_version'):
            self._redis_version = tuple(int(i) for i in self.redis_client.info()['redis_version'].split('.'))

        return self._redis_version

    @property
    def len(self):

        return len(self.events)

    @property
    def queue_key(self):

        return f'{self.key_prefix}:{self.idx}'

    @property
    def lock_idx(self):

        return f'pq:{self.idx}'

    @property
    def _ident(self):
        conn_kwargs = self.redis_client.get_connection_kwargs()
        return f'{conn_kwargs["host"]}:{conn_kwargs["port"]}:{conn_kwargs["db"]}:{self.idx}'

    @property
    def guid(self):
        return hashlib.sha256(self._ident.encode('utf-8')).hexdigest()


class Queue(object):

    key_prefix = 'qm'

    def __init__(self,
                 name,
                 ttl,
                 redis_client,
                 max_retry=DEFAULT_MAX_RETRY):

        self.name = name
        self.ttl = ttl
        self.redis_client = redis_client
        self.results = []
        self.add_delete_queue_lua()
        self.max_retry = max_retry

    def push(self,
             *events):

        if not all(isinstance(event, Event) for event in events):
            raise ValueError('all events need to be of type Event')

        mapping = {k: v for event in events for k, v in event.mapping.items()}
        n = self.redis_client.zadd(self.main_key, mapping=mapping, nx=True)
        self.expire()

        return n

    def push_raw(self,
                 *data):

        events = [Event(d, redis_client=self.redis_client, max_retry=self.max_retry) for d in data]
        return self.push(*events)

    def add_delete_queue_lua(self):

        if not hasattr(self.redis_client, 'delete_queue'):
            self.redis_client.delete_queue = self.redis_client.register_script(LuaScripts.delete_queue())

    def delete(self):

        self.redis_client.delete_queue(
            keys=[
                self.main_key,
                self.register_processing_queue_key
            ]
        )

    def expire(self):

        self.redis_client.expire(self.main_key, self.ttl)

    def get_processing_queue(self,
                             max_elements=None,
                             min_ts=None,
                             max_ts=None,
                             ttl=None):

        return ProcessingQueue.get(
            parent_key=self.main_key,
            register_key=self.register_processing_queue_key,
            main_queue_ttl=self.ttl,
            max_elements=max_elements,
            min_ts=min_ts,
            max_ts=max_ts,
            ttl=ttl,
            redis_client=self.redis_client)

    def get_rogue_processing_queue(self,
                                   margin=120,
                                   offset=0,
                                   ttl=None):

        return ProcessingQueue.get_rogue(
            parent_key=self.main_key,
            register_key=self.register_processing_queue_key,
            main_queue_ttl=self.ttl,
            redis_client=self.redis_client,
            margin=margin,
            offset=offset,
            ttl=ttl)

    def process(self,
                func,
                ptime=120,
                chunk_size=50,
                min_ts='-inf',
                max_ts='+inf',
                handle_rogues=True,
                quit_when_empty=True,
                max_rogues=100,
                interval=1):

        # the func should return a tuple (success, result)
        # the success boolean is used to determine if the event was handled successful

        start = time.time()
        if handle_rogues:
            for offset in range(max_rogues):
                pq = self.get_rogue_processing_queue(offset=offset)
                if pq is None:
                    break
                try:
                    with pq.lock():
                        self.results.append(pq.process(func))
                except (LockError, ConnectionError, TimeoutError):
                    continue

        while time.time() - start <= ptime:
            q = self.get_processing_queue(max_elements=chunk_size, min_ts=min_ts, max_ts=max_ts)
            if not q.len == 0:
                try:
                    with q.lock():
                        self.results.append(q.process(func))
                except (LockError, ConnectionError, TimeoutError):
                    continue
            else:
                if quit_when_empty:
                    break
                else:
                    time.sleep(interval)

        return self.results

    def batch_process(self,
                      func,
                      batch_size,
                      ptime=120,
                      chunk_size=50,
                      min_ts=None,
                      max_ts=None,
                      handle_rogues=True,
                      quit_when_empty=False,
                      max_rogues=100,
                      interval=1):

        # the func should return a tuple (event_id, success, result)
        # the success boolean is used to determine if the event was handled successful

        start = time.time()
        if handle_rogues:
            for offset in range(max_rogues):
                pq = self.get_rogue_processing_queue(offset=offset)
                if pq is None:
                    break
                try:
                    with pq.lock():
                        self.results.extend(pq.batch_process(func, batch_size=batch_size))
                except (LockError, ConnectionError, TimeoutError):
                    continue

        while time.time() - start <= ptime:
            q = self.get_processing_queue(max_elements=chunk_size, min_ts=min_ts, max_ts=max_ts)
            if not q.len == 0:
                try:
                    with q.lock():
                        self.results.extend(q.batch_process(func, batch_size=batch_size))
                except (LockError, ConnectionError, TimeoutError):
                    continue
            else:
                if quit_when_empty:
                    break
                else:
                    time.sleep(interval)

        return self.results

    @property
    def len(self):

        return self.redis_client.zcard(self.main_key)

    @property
    def main_key(self):

        return f'{self.key_prefix}:{self.name}'

    @property
    def register_processing_queue_key(self):

        return f'{self.key_prefix}:{self.name}:processing'

    @property
    def redis_version(self):

        if not hasattr(self, '_redis_version'):
            self._redis_version = tuple(int(i) for i in self.redis_client.info()['redis_version'].split('.'))

        return self._redis_version

    @property
    def _ident(self):
        conn_kwargs = self.redis_client.get_connection_kwargs()
        return f'{conn_kwargs["host"]}:{conn_kwargs["port"]}:{conn_kwargs["db"]}:{self.name}'

    @property
    def guid(self):
        return hashlib.sha256(self._ident.encode('utf-8')).hexdigest()



from requests.sessions import Session
from requests.auth import HTTPBasicAuth
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests import exceptions as requests_exceptions
import urllib3
import pytz
import datetime
import json

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

DEFAULT_HEC_SOURCE = 'python-grains-splunk-client'
DEFAULT_HEC_SOURCE_TYPE = '_json'
DEFAULT_HEC_TIMEOUT = 1.0

class SplunkError(Exception): pass

class SplunkTimeout(SplunkError): pass
class SplunkInvalidConfig(SplunkError): pass
class SplunkInvalidInput(SplunkError): pass

class SplunkClient(object):

    def __init__(self,
                 hostname,
                 scheme='https',
                 management_port=8089,
                 hec_port=8088,
                 hec_token=None,
                 hec_source=None,
                 hec_source_type=None,
                 hec_timeout=None,
                 hec_omit_port=False):

        self.hostname = hostname
        self.scheme = scheme
        self.management_port = management_port
        self.hec_port = hec_port
        self.hec_token = hec_token
        self.hec_omit_port = hec_omit_port
        self._hec_timeout = hec_timeout
        self._hec_source = hec_source
        self._hec_source_type = hec_source_type


    def hec_send(self,
                 events,
                 index,
                 timestamp=None,
                 source=None,
                 sourcetype=None):

        if not isinstance(events, list):
            raise ValueError('events must be a list')

        payloads = [self.splunk_payload(event,
                                        index,
                                        timestamp=timestamp,
                                        source=source,
                                        sourcetype=sourcetype)
                    for event in events]

        return self.hec_send_payloads(payloads)

    def hec_send_payloads(self, payloads):

        payload_strings = [self.splunk_payload_string(payload) for payload in payloads]
        data = ''.join(payload_strings)

        try:
            r = self.http_session.post(self.hec_endpoint,
                                       data=data,
                                       timeout=self.hec_timeout,
                                       headers={'Authorization': self.hec_auth_header})
        except requests_exceptions.Timeout:
            raise SplunkTimeout
        if r.status_code == 200:
            return []
        else:
            return [r.content]

    def splunk_payload(self, event, index, timestamp=None, source=None, sourcetype=None):

        if timestamp is None:
            timestamp = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
        else:
            if not isinstance(timestamp, (float, int)):
                raise ValueError('timestamp must be a float')

        return {
            'time': timestamp,
            'index': index,
            'source': source or self.hec_source,
            'sourcetype': sourcetype or self.hec_sourcetype,
            'event': event
        }

    def splunk_payload_string(self,
                       event):

        return json.dumps(event)

    @property
    def http_session(self):
        if not hasattr(self, '_http_session') or self._http_session is None:
            self._http_session = Session()
            retry_strategy = Retry(
                total=3,
                backoff_factor=1,
                status_forcelist=[429, 500, 502, 503, 504],
                allowed_methods=['GET', 'POST'],
                respect_retry_after_header=True
            )
            self._http_session.mount(self.host, adapter=HTTPAdapter(max_retries=retry_strategy))
        return self._http_session

    @property
    def hec_endpoint(self):
        if self.hec_omit_port:
            return f'{self.host}/services/collector/event'
        return f'{self.host}:{self.hec_port}/services/collector/event'

    @property
    def host(self):
        return f'{self.scheme}://{self.hostname}'

    @property
    def hec_auth_header(self):
        if self.hec_token is None:
            raise SplunkInvalidConfig('No HEC token given')
        return f'Splunk {self.hec_token}'

    @property
    def hec_source(self):
        return self._hec_source or DEFAULT_HEC_SOURCE

    @property
    def hec_sourcetype(self):
        return self._hec_source_type or DEFAULT_HEC_SOURCE_TYPE

    @property
    def hec_timeout(self):
        return self._hec_timeout or DEFAULT_HEC_TIMEOUT


